import importlib
from pathlib import Path
from subprocess import run
from typing import Optional

from lamin_logger import logger

from lndb.test import get_package_name

from .._settings_instance import InstanceSettings
from .._setup_schema import get_schema_module_name


def get_package_info(
    schema_root: Optional[Path] = None, package_name: Optional[str] = None
):
    if package_name is None:
        if schema_root is None:
            raise AssertionError("Must provide schema_root or package_name!")
        package_name = get_package_name(schema_root)
    package = importlib.import_module(package_name)
    if not hasattr(package, "_schema_id"):
        package_name = f"{package_name}.schema"
        package = importlib.import_module(package_name)
    schema_id = getattr(package, "_schema_id")
    migrations_path = Path(package.__file__).parent / "migrations"  # type:ignore
    return package_name, migrations_path, schema_id


def generate_alembic_ini(
    package_name: str,
    migrations_path: Optional[Path] = None,
    schema_id: Optional[str] = None,
):
    if migrations_path is None:
        _, migrations_path, schema_id = get_package_info(package_name=package_name)
    _migrations_path = Path(__file__).parent

    if not (migrations_path.parent / "alembic.ini").exists():
        content = (
            _readfile(_migrations_path / "alembic.ini")
            .replace("[{schema_id}]", f"[{schema_id}]")
            .replace("{package_name}/migrations", f"{package_name}/migrations")
        )
        _writefile(migrations_path.parent / "alembic.ini", content)


def modify_alembic_ini(
    filepath: Path,
    isettings: InstanceSettings,
    schema_name: Optional[str] = None,
    package_name: Optional[str] = None,
    revert: bool = False,
    move_sl: bool = True,
):
    if package_name is None:
        package_name = get_schema_module_name(schema_name)
    schema_module_path = package_name.replace(".", "/") + "/migrations"
    sl_from = schema_module_path
    sl_to = "migrations" if move_sl else schema_module_path
    url_from = "sqlite:///testdb/testdb.lndb"
    url_to_sqlite = f"sqlite:///{isettings._sqlite_file_local}"
    url_to = url_to_sqlite if isettings.dialect == "sqlite" else isettings.db

    if revert:
        sl_from, sl_to = sl_to, sl_from
        url_from, url_to = url_to, url_from

    with open(filepath) as f:
        content = f.read()

    content = content.replace(
        f"script_location = {sl_from}",
        f"script_location = {sl_to}",
    ).replace(
        f"sqlalchemy.url = {url_from}",
        f"sqlalchemy.url = {url_to}",
    )

    with open(filepath, "w") as f:
        f.write(content)


def generate_module_files(
    package_name: str,
    migrations_path: Optional[Path] = None,
    schema_id: Optional[str] = None,
):
    if migrations_path is None:
        _, migrations_path, schema_id = get_package_info(package_name=package_name)
    _migrations_path = Path(__file__).parent

    # ensures migrations/versions folder exists
    (migrations_path / "versions").mkdir(exist_ok=True, parents=True)

    if not (migrations_path / "env.py").exists():
        content = (
            _readfile(_migrations_path / "env.py")
            .replace("_schema_id = None\n", "")
            .replace("# from {package_name} import *", f"from {package_name} import *")
            .replace(
                "# from {package_name} import _schema_id",
                f"from {package_name} import _schema_id",
            )
        )
        _writefile(migrations_path / "env.py", content)

    if not (migrations_path / "script.py.mako").exists():
        import shutil

        shutil.copyfile(
            _migrations_path / "script.py.mako", migrations_path / "script.py.mako"
        )

    generate_alembic_ini(
        package_name=package_name, migrations_path=migrations_path, schema_id=schema_id
    )


def set_alembic_logging_level(migrations_path: Path, level="INFO"):
    alembic_ini_path = migrations_path.parent / "alembic.ini"
    text = _readfile(alembic_ini_path)
    current_level = text.split("[logger_alembic]\n")[1].split("\n")[0]
    new_level = f"level = {level}"
    _writefile(
        alembic_ini_path,
        text.replace(
            f"[logger_alembic]\n{current_level}", f"[logger_alembic]\n{new_level}"
        ),
    )


def _readfile(filename: Path):
    with open(filename, "r") as f:
        return f.read()


def _writefile(filename: Path, content: str):
    with open(filename, "w") as f:
        return f.write(content)


push_instruction = """Please push your changes to a new remote branch and open a PR.
Inspect CI step Build output and add migration code to the script.
It will look like the following, beware of the renaming:
    op.drop_column('biosample', 'description', schema='wetlab')
    op.add_column('experiment', sa.Column('donor_id', sqlmodel.sql.sqltypes.AutoString(), schema="wetlab")"""  # noqa


class migrate:
    """Manage migrations."""

    @staticmethod
    def generate(
        version: str = "vX.X.X",
        schema_root: Optional[Path] = None,
        package_name: Optional[str] = None,
    ):
        """Generate migration for current schema module.

        Needs to be executed at the root level of the python package that contains
        the schema module.

        Args:
            version: Version string to label migration with.
            schema_root: Optional. Root directory of schema module.
            package_name: Optional. Name of schema module package.
        """
        package_name, migrations_path, schema_id = get_package_info(
            schema_root=schema_root, package_name=package_name
        )
        generate_module_files(
            package_name=package_name,  # type:ignore
            migrations_path=migrations_path,
            schema_id=schema_id,
        )
        testdb_path = migrations_path.parent.parent / "testdb/testdb.lndb"  # type: ignore # noqa
        if testdb_path.exists():
            # runs dev mode to write migration scripts
            rm = False
            set_alembic_logging_level(migrations_path, level="INFO")
            logger.info("Generate migration with reference db: testdb/testdb.lndb")
        else:
            # runs CI-guided mode to generate empty migration scripts
            rm = True
            from lndb._settings import settings

            modify_alembic_ini(
                filepath=migrations_path.parent / "alembic.ini",
                isettings=settings.instance,
                package_name=package_name,
                move_sl=False,
            )
            set_alembic_logging_level(migrations_path, level="WARN")
            logger.info("Generate empty migration script.")
        command = (
            f"alembic --config {package_name}/alembic.ini --name {schema_id} revision"
            f" --autogenerate -m '{version}'"
        )
        if schema_root is not None:
            cwd = f"{schema_root}"
        else:
            cwd = None
        process = run(command, shell=True, cwd=cwd)

        if process.returncode == 0:
            logger.success(f"Successfully generated migration {version}.")
            if rm:
                modify_alembic_ini(
                    filepath=migrations_path.parent / "alembic.ini",
                    isettings=settings.instance,
                    package_name=package_name,
                    move_sl=False,
                    revert=True,
                )
                logger.info(push_instruction)
            return None
        else:
            print(process.stderr)
            logger.error("Generating migration failed.")
            return "migrate-gen-failed"



# We see a lot of import statements for lnschema_core below
# This is currently needed as we can only import the schema module
# once isettings have been adjusted
from typing import Any, List

import sqlmodel as sqm
from lamin_logger import logger
from sqlalchemy.exc import ProgrammingError

from ._settings import settings


class upsert:
    @classmethod
    def user(cls, email: str, user_id: str, handle: str, name: str = None):
        import lnschema_core as schema_core

        with sqm.Session(settings.instance.engine) as session:
            user_table = (
                schema_core.User if hasattr(schema_core, "User") else schema_core.user
            )  # noqa
            try:
                user = session.get(user_table, user_id)
            except ProgrammingError as e:
                logger.error(
                    "Cannot find user table.\nLikely, you try to run SQLite and"
                    " Postgres from the same process.\n\n->Abort and restart with only"
                    " one SQL dialect in one process.\n\n"
                )
                raise e
        if user is None:
            user_id = insert.user(email, user_id, handle, name)  # type: ignore
            # do not update sqlite on the cloud as this happens within
            # insert.user
        else:
            # update the user record
            update_email = email != user.email
            update_handle = handle != user.handle
            # name = None is currently understood as no user name yet provided
            update_name = name != user.name and name is not None

            if any((update_email, update_handle, update_name)):
                with sqm.Session(settings.instance.engine) as session:
                    msg = "Updating: "
                    if update_email:
                        msg += f"{user.email} -> {email} "
                        user.email = email
                    if update_handle:
                        msg += f"{user.handle} -> {handle} "
                        user.handle = handle
                    if update_name:
                        msg += f"{user.name} -> {name} "
                        user.name = name
                    logger.info(msg)
                    session.add(user)
                    session.commit()

                settings.instance._update_cloud_sqlite_file()

        return user_id


class insert_if_not_exists:
    """Insert data if it does not yet exist.

    A wrapper around the `insert` class below.
    """

    @classmethod
    def storage(cls, root, region):
        import lnschema_core as schema_core

        root = str(root)
        with sqm.Session(settings.instance.engine) as session:
            storage_table = (
                schema_core.Storage
                if hasattr(schema_core, "Storage")
                else schema_core.storage
            )  # noqa
            storage = session.exec(
                sqm.select(storage_table).where(storage_table.root == root)
            ).first()
        if storage is None:
            storage = insert.storage(root, region)  # type: ignore

        return storage


class insert:
    """Insert data."""

    @classmethod
    def version(
        cls,
        *,
        schema_module: Any,
        user_id: str,
        cloud_sqlite: bool = True,
    ):
        """Core schema module version.

        Args:
            schema_module: The schema module.
            user_id: User ID.
            cloud_sqlite: Update cloud SQLite file or not.
        """
        schema_id, version, migration = (
            schema_module._schema_id,
            schema_module.__version__,
            schema_module._migration,
        )

        with sqm.Session(settings.instance.engine) as session:
            table_loc = (
                schema_module.dev if hasattr(schema_module, "dev") else schema_module
            )
            version_table = getattr(table_loc, f"version_{schema_id}")
            session.add(version_table(v=version, migration=migration, user_id=user_id))
            session.commit()

        if cloud_sqlite:
            settings.instance._update_cloud_sqlite_file()

    @classmethod
    def user(cls, email, user_id, handle, name):
        """User."""
        import lnschema_core as schema_core

        with sqm.Session(settings.instance.engine) as session:
            user_table = (
                schema_core.User if hasattr(schema_core, "User") else schema_core.user
            )  # noqa
            user = user_table(id=user_id, email=email, handle=handle, name=name)
            session.add(user)
            session.commit()
            session.refresh(user)

        settings.instance._update_cloud_sqlite_file()

        return user.id

    @classmethod
    def storage(cls, root, region):
        """Storage."""
        import lnschema_core as schema_core

        with sqm.Session(settings.instance.engine) as session:
            storage_table = (
                schema_core.Storage
                if hasattr(schema_core, "Storage")
                else schema_core.storage
            )  # noqa
            storage = storage_table(
                root=root, region=region, type=settings.instance.storage.type
            )
            session.add(storage)
            session.commit()
            session.refresh(storage)

        settings.instance._update_cloud_sqlite_file()

        return storage

    @classmethod
    def bionty_versions(cls, records: List[sqm.SQLModel]):
        """Bionty versions."""
        from lnschema_bionty import dev

        with sqm.Session(settings.instance.engine) as session:
            for record in records:
                session.add(record)
                current_record = dev.CurrentBiontyVersions(
                    id=record.id, entity=record.entity
                )
                session.add(current_record)
            session.commit()

        settings.instance._update_cloud_sqlite_file()



import pytest

import lndb as lndb


def test_db_unique_error():
    lndb.login("testuser2")

    # postgres
    with pytest.raises(RuntimeError):
        lndb.init(
            storage="s3://lndb-setup-ci",
            schema="retro,bionty",
            db="postgresql://batman:robin@35.222.187.204:5432/retro",
        )

    # sqlite
    # this fails because there is already an sqlite with the same name in that bucket
    # hence, the sqlite file would clash

    # with pytest.raises(RuntimeError):
    #     lndb.init(storage="s3://lamindb-ci")


def test_value_error_schema():
    with pytest.raises(ValueError):
        lndb.init(storage="tmpstorage1", schema="bionty, xyz1")



from textwrap import dedent


def doc_args(*args):
    """Pass arguments to docstrings."""

    def dec(obj):
        obj.__orig_doc__ = obj.__doc__
        obj.__doc__ = dedent(obj.__doc__).format(*args)
        return obj

    return dec



from dataclasses import dataclass
from typing import Union


class user_description:
    email = """User email."""
    password = """User password."""
    id = """User ID."""
    handle = "Unique handle."
    name = "Full name."


@dataclass
class UserSettings:
    """User data. All synched from cloud."""

    email: str = None  # type: ignore
    """User email."""
    password: Union[str, None] = None
    """User password."""
    access_token: Union[str, None] = None
    """User access token."""
    id: Union[str, None] = None
    """User ID."""
    handle: Union[str, None] = None
    "Unique handle."
    name: Union[str, None] = None
    "Full name."

    def __repr__(self):
        """Rich string representation."""
        representation = f"Current user: {self.handle}"
        attrs = ["handle", "email", "id"]
        for attr in attrs:
            value = getattr(self, attr)
            representation += f"\n- {attr}: {value}"
        return representation



from pathlib import Path
from typing import Literal, Optional, Union

from appdirs import AppDirs
from cloudpathlib import CloudPath, GSClient, S3Client
from cloudpathlib.exceptions import OverwriteNewerLocalError
from lamin_logger import logger

DIRS = AppDirs("lamindb", "laminlabs")


_MUTE_SYNC_WARNINGS = False


def _set_mute_sync_warnings(value: bool):
    global _MUTE_SYNC_WARNINGS

    _MUTE_SYNC_WARNINGS = value


class Storage:
    """Manage cloud or local storage."""

    def __init__(self, root: Union[str, Path, CloudPath], region: Optional[str] = None):
        if isinstance(root, str):
            root_path = Storage._str_to_path(root)
        else:
            root_path = root
        self._root = root_path
        self._region = region

    @staticmethod
    def _str_to_path(storage: str) -> Union[Path, CloudPath]:
        if str(storage).startswith("s3://"):  # AWS
            storage_root = CloudPath(storage)
        elif str(storage).startswith("gs://"):  # GCP
            # the below seems needed as cloudpathlib on its
            # own fails to initialize when using gcloud auth login
            # and not JSON credentials
            from cloudpathlib import GSClient
            from google.cloud import storage as gstorage

            client = GSClient(storage_client=gstorage.Client())
            storage_root = CloudPath(storage, client)
        else:  # local path
            storage_root = Path(storage).absolute()
        return storage_root

    @property
    def root(self) -> Union[Path, CloudPath]:
        """Root storage location."""
        return self._root

    @property
    def cache_dir(
        self,
    ) -> Union[Path, None]:
        """Cache root, a local directory to cache cloud files."""
        if self.is_cloud:
            cache_dir = Path(DIRS.user_cache_dir)
            cache_dir.mkdir(parents=True, exist_ok=True)
        else:
            cache_dir = None
        return cache_dir

    @property
    def is_cloud(self) -> bool:
        """`True` if `storage_root` is in cloud, `False` otherwise."""
        return isinstance(self.root, CloudPath)

    @property
    def region(self) -> Optional[str]:
        """Storage region."""
        return self._region

    @property
    def client(self) -> Union[S3Client, GSClient]:
        if self.type == "s3":
            return S3Client(local_cache_dir=self.cache_dir)
        elif self.type == "gs":
            # the below seems needed as cloudpathlib on its
            # own fails to initialize when using gcloud auth login
            # and not JSON credentials
            from google.cloud import storage as gstorage

            return GSClient(
                local_cache_dir=self.cache_dir,
                storage_client=gstorage.Client(),
            )
        elif self.type == "local":
            raise RuntimeError("You shouldn't need a client for local storage.")
        else:
            raise RuntimeError(
                "Currently, only AWS S3 and Google cloud are supported for cloud"
                " storage."
            )

    @property
    def type(self) -> Literal["s3", "gs", "local"]:
        """AWS S3 vs. Google Cloud vs. local.

        Returns "s3" or "gs" or "local".
        """
        if str(self.root).startswith("s3://"):
            return "s3"
        elif str(self.root).startswith("gs://"):
            return "gs"
        else:
            return "local"

    def key_to_filepath(
        self, filekey: Union[Path, CloudPath, str]
    ) -> Union[Path, CloudPath]:
        """Cloud or local filepath from filekey."""
        if self.is_cloud:
            return self.client.CloudPath(self.root / filekey)
        else:
            return self.root / filekey

    def cloud_to_local(self, filepath: Union[Path, CloudPath]) -> Path:
        """Local (cache) filepath from filepath."""
        try:
            # the following will auto-update the local cache if the cloud file is newer
            # if both have the same age, it will keep it as is
            if self.is_cloud:
                local_filepath = self.client.CloudPath(filepath).fspath
            else:
                local_filepath = Path(filepath)
        except OverwriteNewerLocalError:
            local_filepath = self.cloud_to_local_no_update(filepath)  # type: ignore
            if not _MUTE_SYNC_WARNINGS:
                logger.warning(
                    f"Local file ({local_filepath}) for cloud path ({filepath}) is newer on disk than in cloud.\n"  # noqa
                    "It seems you manually updated the database locally and didn't push changes to the cloud.\n"  # noqa
                    "This can lead to data loss if somebody else modified the cloud file in"  # noqa
                    " the meantime."
                )
        Path(local_filepath).parent.mkdir(
            parents=True, exist_ok=True
        )  # this should not happen here but is currently needed
        return local_filepath

    # conversion to Path via cloud_to_local() would trigger download
    # of remote file to cache if there already is one
    # in pure write operations that update the cloud, we don't want this
    # hence, we manually construct the local file path
    # using the `.parts` attribute in the following line
    def cloud_to_local_no_update(self, filepath: Union[Path, CloudPath]) -> Path:
        if self.is_cloud:
            return self.cache_dir.joinpath(*filepath.parts[1:])  # type: ignore
        return filepath

    def local_filepath(self, filekey: Union[Path, CloudPath, str]) -> Path:
        """Local (cache) filepath from filekey: `local(filepath(...))`."""
        return self.cloud_to_local(self.key_to_filepath(filekey))



from pathlib import Path
from typing import Optional

import yaml  # type: ignore


def get_package_name(root_directory: Optional[Path] = None) -> str:
    yaml_file = Path("lamin-project.yaml")
    if root_directory is None:
        root_directory = Path(".")
    yaml_file = root_directory / yaml_file
    with yaml_file.open() as f:
        package_name = yaml.safe_load(f)["package_name"]
    return package_name



from pathlib import Path
from typing import Optional, Union

from cloudpathlib import CloudPath
from lamin_logger import logger
from lnhub_rest._add_storage import add_storage as add_storage_hub

from ._init_instance import register
from ._settings import settings
from ._settings_instance import InstanceSettings
from ._settings_load import load_instance_settings
from ._settings_store import current_instance_settings_file, instance_settings_file


def set_storage(
    storage: Union[str, Path, CloudPath], instance_name: Optional[str] = None
):
    """Set storage."""
    if settings.instance.owner != settings.user.handle:
        logger.error("Can only set storage if current user is instance owner.")
        return "only-owner-can-set-storage"
    settings_file = (
        instance_settings_file(instance_name, settings.instance.owner)
        if instance_name
        else current_instance_settings_file()
    )
    isettings = load_instance_settings(settings_file)

    if isettings.dialect == "sqlite":
        logger.error("Can't set storage for sqlite instance.")
        return "set-storage-failed"

    new_isettings = InstanceSettings(
        owner=isettings.owner,
        name=isettings.name,
        storage_root=storage,
        db=isettings.db,
        schema=isettings._schema_str,
    )

    new_isettings._persist()
    register(new_isettings, settings.user)
    logger.info(
        f"Set storage {storage} for instance {isettings.owner}/{isettings.name}"
    )
    if isettings.is_remote:
        add_storage_hub(storage, account_handle=isettings.owner)
